package org.processmining.models.queueminer.distribution;

import java.util.ArrayList;
import java.util.HashMap;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.models.queueminer.patternvisualizer.DistributionInfo;
import org.processmining.models.queueminer.patternvisualizer.GranularityFindingUtils;
import org.processmining.models.queueminer.utils.TimeActivity;
import org.processmining.models.queueminer.utils.TimeActivityUtilities;

public class DistributionFilter {
	
	/**
	 * Finds all activity types of a log
	 * @param log
	 * @return
	 */
	public static ArrayList<String> getActivityTypes(XLog log){
		ArrayList<String> actTypes = new ArrayList<>();
		XConceptExtension xce = XConceptExtension.instance();
		for(XTrace t:log){
			for(int i = 0; i < t.size(); i++){
				String type = xce.extractName(t.get(i));
				if(!actTypes.contains(type)){
					actTypes.add(type);
				}
			}
		}
		return actTypes;
	}
	
	public static ArrayList<DistributionInfo> splitRootDist(DistributionInfo dinf, int nrParts, int threshold, long intervalTime, long weekStart){
		if(!dinf.approved){
			System.err.println("Trying to split disapproved distribution!");
			return null;
		}
		
		ArrayList<TimeActivity> tlog = dinf.tal;
		long partSize = intervalTime/nrParts;
		
		ArrayList<DistributionInfo> splitDists = new ArrayList<>();
		HashMap<Integer, ArrayList<TimeActivity>> splitTals = new HashMap<>();
		for(int i = 0; i < nrParts; i++){
			splitTals.put(i, new ArrayList<TimeActivity>());
		}
		//Put all the activities in the correct subpart
		for(TimeActivity ta:tlog){
			long taTime = (ta.start-weekStart) % intervalTime;
			for(int i = 0; i < nrParts;i++){
				long pstart = i*partSize;
				long pend = pstart+partSize;
				if(taTime >= pstart && taTime <= pend){
					splitTals.get(i).add(ta);
				}
			}
		}
		
		//Then create new splitDists
		for(int i = 0; i < nrParts; i++){
			ArrayList<TimeActivity> splitLog = splitTals.get(i);
			ArrayList<Point> splitDist = DistributionFilter.findDistribution(splitLog, partSize);
			int splitDistSize = GranularityFindingUtils.findPointListSize(splitDist);
			boolean approved = (splitDistSize >= threshold);
			long pstart = i*partSize;
			long pend = pstart+partSize;
			splitDists.add(new DistributionInfo(splitDist,splitLog,approved,pstart,pend));
		}
		
		return splitDists;
	}
	
	public static ArrayList<DistributionInfo> splitDistributionInfo(DistributionInfo dinf, int nrParts, int threshold, long intervalTime, long binSize, long weekStart){
		if(!dinf.approved){
			System.err.println("Trying to split disapproved distribution!");
			return null;
		}
		
		ArrayList<TimeActivity> tlog = dinf.tal;
		long start = dinf.start;
		long end = dinf.end;
		long partSize = (end-start)/nrParts;
		
		ArrayList<DistributionInfo> splitDists = new ArrayList<>();
		HashMap<Integer, ArrayList<TimeActivity>> splitTals = new HashMap<>();
		for(int i = 0; i < nrParts; i++){
			splitTals.put(i, new ArrayList<TimeActivity>());
		}
		//Put all the activities in the correct subpart
		for(TimeActivity ta:tlog){
			long taTime = (ta.start-weekStart) % intervalTime;
			for(int i = 0; i < nrParts;i++){
				long pstart = start+i*partSize;
				long pend = pstart+partSize;
				if(taTime >= pstart && taTime <= pend){
					splitTals.get(i).add(ta);
				}
			}
		}
		
		//Then create new splitDists
		for(int i = 0; i < nrParts; i++){
			ArrayList<TimeActivity> splitLog = splitTals.get(i);
			ArrayList<Point> splitDist = DistributionFilter.findDistribution(splitLog, binSize);
			int splitDistSize = GranularityFindingUtils.findPointListSize(splitDist);
			boolean approved = (splitDistSize >= threshold);
			long pstart = start+i*partSize;
			long pend = pstart+partSize;
			splitDists.add(new DistributionInfo(splitDist,splitLog,approved,pstart,pend));
		}
		
		return splitDists;
	}
	
	public static ArrayList<Point> findDistribution(ArrayList<TimeActivity> tlog, long binSize){
		long minTime = logMinDuration(tlog);
		long maxTime = logMaxDuration(tlog);
		long nrBins = ((maxTime-minTime)/binSize)+1;
		int[] binCount = new int[(int) nrBins];
		for(TimeActivity ta:tlog){
			long duration = ta.end-ta.start;
			long bin = (duration-minTime)/binSize;
			binCount[(int) bin]++;
		}
		return binsToPointList(binCount,minTime,binSize);
	}
	
	/**
	 * 
	 * @param log
	 * @param actType
	 * @param binSize
	 * @return
	 */
	public static ArrayList<Point> activityTypeFilterDistribution(XLog log, String actType, long binSize){
		ArrayList<TimeActivity> tlog = TimeActivityUtilities.findTimeActivities(log);
		
		long minTime = logMinDuration(tlog);
		long maxTime = logMaxDuration(tlog);
		long nrBins = ((maxTime-minTime)/binSize)+1;
		//find all activities of that type
		ArrayList<TimeActivity> spectlog = new ArrayList<>();
		for(TimeActivity ta:tlog){
			if(ta.type.equals(actType)){
				spectlog.add(ta);
			}
		}
		//find the corresponding bins
		int[] binCount = new int[(int) nrBins];
		for(TimeActivity ta:spectlog){
			long duration = ta.end-ta.start;
			long bin = (duration-minTime)/binSize;
			binCount[(int) bin]++;
		}
		return binsToPointList(binCount,minTime,binSize);
	}
	
	/**
	 * Filter based on workpressure, where workpressure = #running cases
	 * @param log
	 * @param binSize
	 * @param intervalStart
	 * @param intervalEnd
	 * @param intervalSpacing
	 * @return
	 */
	public static ArrayList<Point> workPressureFilterDistribution(XLog log, long binSize, int workPressureMin, int workPressureMax, 
			long startTime, long subTime, int nrSubs){
		ArrayList<TimeActivity> tlog = TimeActivityUtilities.findTimeActivities(log);
		long minTime = logMinDuration(tlog);
		long maxTime = logMaxDuration(tlog);
		long nrBins = ((maxTime-minTime)/binSize)+1;
		
		//First find all bins with the right workpressure
		ArrayList<TimeActivity> workTLog = new ArrayList<>();
		int workPressure;
		ArrayList<TimeActivity> cur = new ArrayList<>();
		int maxte = 0;
		for(int i = 0; i < nrSubs; i++){
			long substart = startTime + subTime * i;
			long subend = substart + subTime;
			//Find the activities that are still relevant
			ArrayList<TimeActivity> ncur = new ArrayList<>();
			for(int j = 0; j < cur.size(); j++){
				TimeActivity ta = cur.get(j);
				if(ta.end >= substart){
					ncur.add(ta);
				}
			}
			//dan daarna checken tot de start voorbij current sub is
			for(int j = maxte; j < tlog.size(); j++){
				TimeActivity te = tlog.get(j);
				if(te.start > subend){
					break;
				}
				else{
					ncur.add(te);
					maxte++;
				}
			}
			cur = ncur;
			workPressure = cur.size();
			if(workPressure >= workPressureMin && workPressure <= workPressureMax){
				workTLog.addAll(cur);
			}
		}
		
		//Then bin all the activities with the right workpressure
		int[] binCount = new int[(int) nrBins];
		for(TimeActivity ta:workTLog){
			long duration = ta.end-ta.start;
			long bin = (duration-minTime)/binSize;
			binCount[(int) bin]++;
		}
		return binsToPointList(binCount,minTime,binSize);
	}
	
	/**
	 * @pre end > start
	 * @param log
	 * @param binSize
	 * @param intervalStart
	 * @param intervalEnd
	 * @param intervalSpacing
	 * @return
	 */
	public static ArrayList<Point> TimeFilterDistribution(XLog log, long binSize, long intervalStart, long intervalEnd, long intervalSpacing){
		ArrayList<TimeActivity> tlog = TimeActivityUtilities.findTimeActivities(log);
		long minTime = logMinDuration(tlog);
		long maxTime = logMaxDuration(tlog);
		long nrBins = ((maxTime-minTime)/binSize)+1;
		int[] binCount = new int[(int) nrBins];
		for(TimeActivity ta:tlog){
			if(ta.start % intervalSpacing >= intervalStart && ta.start % intervalSpacing <= intervalEnd){
				long duration = ta.end-ta.start;
				long bin = (duration-minTime)/binSize;
				binCount[(int) bin]++;
			}
		}
		return binsToPointList(binCount,minTime,binSize);
	}
	
	public static ArrayList<Point> binsToPointList(int[] binCount, long minTime, long binSize){
		//Remove trailing empty bins
		int lastValue = 0;
		for(int i = 0; i < binCount.length; i++){
			if(binCount[i] > 0 ){
				lastValue = i;
			}
		}
		
		ArrayList<Point> result = new ArrayList<>();
		for(int i = 0; i <= lastValue;i++){
			result.add(new Point((minTime+i*binSize),binCount[i]));
		}
		return result;
	}
	
	public static ArrayList<Point> WholeLogDistribution(XLog log, long binSize){
		ArrayList<TimeActivity> tlog = TimeActivityUtilities.findTimeActivities(log);
		long minTime = logMinDuration(tlog);
		long maxTime = logMaxDuration(tlog);
		long nrBins = ((maxTime-minTime)/binSize)+1;
		
		long[] binCount = new long[(int) nrBins];
		for(TimeActivity ta:tlog){
			long duration = ta.end-ta.start;
			long bin = (duration-minTime)/binSize;
			binCount[(int) bin]++;
		}
		
		ArrayList<Point> result = new ArrayList<>();
		for(int i = 0; i < binCount.length;i++){
			result.add(new Point((minTime+i*binSize),binCount[i]));
		}
		return result;
	}
	
	public static long logMinTime(ArrayList<TimeActivity> log){
		long min = Long.MAX_VALUE;
		for(TimeActivity ta:log){
			long time = ta.start;
			if(time < min){
				min = time;
			}
		}
		return min;
	}
	
	public static long logMinDuration(ArrayList<TimeActivity> log){
		long min = Long.MAX_VALUE;
		for(TimeActivity ta:log){
			long duration = ta.end-ta.start;
			if(duration < min){
				min = duration;
			}
		}
		return min;
	}
	
	public static long logMaxDuration(ArrayList<TimeActivity> log){
		long max = Long.MIN_VALUE;
		for(TimeActivity ta:log){
			long duration = ta.end-ta.start;
			if(duration > max){
				max = duration;
			}
		}
		return max;
	}
}
