package org.processmining.models.queueminer.utils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.models.queueminer.distribution.DistributionFilter;
import org.processmining.models.queueminer.distribution.Point;

public class TimeActivityUtilities {
	
	public static ArrayList<Point> graphActivityList(ArrayList<TimeActivity> tal, long subTime){
		long minTime = DistributionFilter.logMinDuration(tal);
		long maxTime = DistributionFilter.logMaxDuration(tal);
		long nrBins = ((maxTime-minTime)/subTime)+1;
		int[] binCount = new int[(int) nrBins];
		for(TimeActivity ta:tal){
			long duration = ta.end-ta.start;
			long bin = (duration-minTime)/subTime;
			binCount[(int) bin]++;
		}
		return DistributionFilter.binsToPointList(binCount,minTime,subTime);
	}
	
	/**
	 * Finds all activities that are in a given time span (subPart * subTime) until ((subPart+1) * subTime)
	 * @param whole
	 * @param subTime
	 * @param subPart
	 * @param intervalTime
	 * @return
	 */
	public static ArrayList<TimeActivity> findSubPartActivities(ArrayList<TimeActivity> whole, long subTime, int subPart, long intervalTime){
		ArrayList<TimeActivity> result = new ArrayList<>();
		long subStart = subTime * subPart;
		long subEnd = subTime * (subPart+1);
		long minTime = DistributionFilter.logMinDuration(whole);
		for(TimeActivity ta: whole){
			long realTime = (ta.start - minTime)%intervalTime;
			//System.out.println("substart: "+subStart+", subend: "+subEnd+", real: "+realTime);
			if(realTime >= subStart && realTime <= subEnd){
				
				result.add(ta);
			}
		}
		return result;
	}

	/**
	 * Finds a list of timed events, given some log
	 * @param log
	 * @return
	 */
	public static ArrayList<TimeActivity> findTimeActivities(XLog log){
		ArrayList<TimeActivity> result = new ArrayList<>();
		XConceptExtension xce = XConceptExtension.instance();
		XTimeExtension te = XTimeExtension.instance();
		int emptyC = 0;
		for(XTrace t:log){
			if(t.size() > 0){
				for(int i = 0; i < t.size()-1; i++){
					long start = te.extractTimestamp(t.get(i)).getTime();
					long end = te.extractTimestamp(t.get(i+1)).getTime();
					String type = xce.extractName(t.get(i));
					result.add(new TimeActivity(start,end,type,t.get(i)));
				}
				long start = te.extractTimestamp(t.get(t.size()-1)).getTime();
				long end = start;
				String type = xce.extractName(t.get(t.size()-1));
				result.add(new TimeActivity(start,end,type,t.get(t.size()-1)));
			}
			else{
				emptyC++;
			}
		}
		//System.out.println("empty traces: "+emptyC);
		return result;
	}
	
	/**
	 * Finds all the activities belonging to queue q and groups them
	 * @param log
	 * @param q
	 * @return
	 */
	public static ArrayList<TimeActivity> findQueueActivities(XLog log, List<String> activities){
		ArrayList<TimeActivity> tal = new ArrayList<>();
		List<String> actNames = activities;
		XConceptExtension xce = XConceptExtension.instance();
		XTimeExtension te = XTimeExtension.instance();
		for(XTrace t:log){
			for(int i = 0; i < t.size()-1; i++){
				String type = xce.extractName(t.get(i));
				if(actNames.contains(type)){
					long start = te.extractTimestamp(t.get(i)).getTime();
					long end = te.extractTimestamp(t.get(i+1)).getTime();
					tal.add(new TimeActivity(start,end,type,t.get(i)));
				}
			}
		}
		return tal;
	}
	
	
	/**
	 * Finds timeactivities and bins them according to interval and subtime
	 * @param log
	 * @param intervalTime
	 * @param subTime
	 * @return
	 */
	public static HashMap<Integer,ArrayList<TimeActivity>> findBinnedTimeActivites(XLog log, long intervalTime, long subTime){
		HashMap<Integer,ArrayList<TimeActivity>> binnedTimeActivities = new HashMap<Integer,ArrayList<TimeActivity>>();
		int nrBins = (int) (intervalTime/subTime)+1;
		//initialize empty hashmap
		for(int i = 0; i < nrBins; i++){
			binnedTimeActivities.put(i, new ArrayList<TimeActivity>());
		}
		XConceptExtension xce = XConceptExtension.instance();
		XTimeExtension te = XTimeExtension.instance();
		//fill hashmap arraylists
		for(XTrace t:log){
			for(int i = 0; i < t.size()-1; i++){
				long start = te.extractTimestamp(t.get(i)).getTime();
				long end = te.extractTimestamp(t.get(i+1)).getTime();
				String type = xce.extractName(t.get(i));
				int bin = (int) ((start % intervalTime)/subTime);
				binnedTimeActivities.get(bin).add(new TimeActivity(start,end,type,t.get(i)));
			}
		}
		return binnedTimeActivities;
	}
	
	public static HashMap<List<String>,HashMap<Integer,ArrayList<TimeActivity>>> findBinnedActivityTimes(XLog log, long intervalTime, long subTime, ArrayList<Queue> queues){
		HashMap<List<String>,HashMap<Integer,ArrayList<TimeActivity>>> result = new HashMap<List<String>,HashMap<Integer,ArrayList<TimeActivity>>>();
		int nrBins = (int) (intervalTime/subTime)+1;
		//build empty datastructure
		for(Queue q:queues){ 
			HashMap<Integer,ArrayList<TimeActivity>> inter  = new HashMap<Integer,ArrayList<TimeActivity>>();
			for(int i = 0; i < nrBins; i++){
				inter.put(i, new ArrayList<TimeActivity>());
			}
			result.put(q.getActivities(), inter);
		}
		//Fill the result datastructure with data
		XConceptExtension xce = XConceptExtension.instance();
		XTimeExtension te = XTimeExtension.instance();
		for(XTrace t:log){
			for(int i = 0; i < t.size()-1; i++){
				long start = te.extractTimestamp(t.get(i)).getTime();
				long end = te.extractTimestamp(t.get(i+1)).getTime();
				String type = xce.extractName(t.get(i));
				int bin = (int) ((start % intervalTime)/subTime);
				for(List<String> str:result.keySet()){
					if(str.contains(type)){
						result.get(str).get(bin).add(new TimeActivity(start,end,type,t.get(i)));
					}
				}
			}
		}
		
		return result;
	}
	
	
	public static HashMap<List<String>,HashMap<Integer,ArrayList<Point>>> graphDataAct(HashMap<List<String>, HashMap<Integer, ArrayList<TimeActivity>>> data, long subTime){
		HashMap<List<String>,HashMap<Integer,ArrayList<Point>>> result = new HashMap<List<String>,HashMap<Integer,ArrayList<Point>>>();
		//add string stuff
		for(List<String> s:data.keySet()){
			HashMap<Integer,ArrayList<Point>> respart = new HashMap<Integer,ArrayList<Point>>();
			result.put(s,respart);
		}
		
		//find points
		long binSize = subTime;
		for(List<String> s:data.keySet()){
			for(Integer i:data.get(s).keySet()){
				ArrayList<TimeActivity> tlog = data.get(s).get(i);
				long minTime = DistributionFilter.logMinDuration(tlog);
				long maxTime = DistributionFilter.logMaxDuration(tlog);
				long nrBins = ((maxTime-minTime)/binSize)+1;
				int[] binCount = new int[(int) nrBins];
				for(TimeActivity ta:tlog){
					long duration = ta.end-ta.start;
					long bin = (duration-minTime)/binSize;
					binCount[(int) bin]++;
				}
				result.get(s).put(i,DistributionFilter.binsToPointList(binCount,minTime,binSize));
			}
		}
		return result;
	}
	
	/**
	 * Return map of pointsets for all distributions in map
	 * @param data
	 * @param subTime
	 * @return
	 */
	public static HashMap<Integer,ArrayList<Point>> graphData(HashMap<Integer,ArrayList<TimeActivity>> data, long subTime){
		HashMap<Integer,ArrayList<Point>> result = new HashMap<Integer,ArrayList<Point>>();
		for(Integer i:data.keySet()){
			long binSize = subTime;
			ArrayList<TimeActivity> tlog = data.get(i);
			long minTime = DistributionFilter.logMinDuration(tlog);
			long maxTime = DistributionFilter.logMaxDuration(tlog);
			long nrBins = ((maxTime-minTime)/binSize)+1;
			int[] binCount = new int[(int) nrBins];
			for(TimeActivity ta:tlog){
				long duration = ta.end-ta.start;
				long bin = (duration-minTime)/binSize;
				binCount[(int) bin]++;
			}
			result.put(i,DistributionFilter.binsToPointList(binCount,minTime,binSize));
		}
		return result;
	}

}
